'''
    TweeLoc is created by

    Pavlos Paraskevopoulos:
        George Mason University
        LIPADE, Paris Descartes University
        pparaske@gmu.edu

    Giovanni Pellegrini:
        University of Trento
        giovanni.pellegrini@unitn.it

    Themis Palpanas:
        Paris Descartes University
        themis@mi.parisdescartes.fr

    Copyright Paraskevopoulos, Pellegrini, Palpanas.


    This software cannot be used for commercial purposes.

    You may use the software as is, only for research purposes.
    You should acknowledge the authors by referencing the following two papers (the first describes the method and algorithms; the second describes the system and visualizations):

    Pavlos Paraskevopoulos, Themis Palpanas.
    Where has this Tweet Come From? Fast and Fine-Grained Geolocalisation of Non-Geotagged Tweets. Social Network Analysis and Mining (SNAM) Journal 6(1), pp 1-16, 2016.
    http://www.mi.parisdescartes.fr/~themisp/publications/snam16.pdf

    Pavlos Paraskevopoulos, Giovanni Pellegrini, Themis Palpanas. TweeLoc: A System for Geolocalizing Tweets at Fine-Grain. International Conference on Data Mining (ICDM), New Orleans, LA, USA, November 2017.
    http://www.mi.parisdescartes.fr/~themisp/publications/icdm17-tweeloc.pdf
'''

import geoplotlib
from geoplotlib.layers import *
from threading import Thread
from PySide.QtGui import *
from PySide.QtCore import QCoreApplication
import sys
import utilities
import glob
import os


class TweetList(QListWidget):
    def __init__(self):
        QListWidget.__init__(self)
        self.itemDoubleClicked.connect(self.show_tweet)
        self.tweets = []

    def show_tweet(self, item):
        text = str(item.text())
        s_text = text.split("Id: ")
        tweet_data = None
        for t in self.tweets:
            if t.id == s_text[1]:
                tweet_data = t
        dialog = TweetDialog()
        dialog.show_tweet(tweet_data)
        dialog.exec_()

    def update_tweets_list(self, tweet):
        if len(self.tweets) < 10:
            self.tweets.append(tweet)
        else:
            del self.tweets[0]
            self.tweets.append(tweet)

    def get_tweet_list(self):
        return self.tweets


class TweetDialog(QDialog):
    def __init__(self, parent=None):
        super(TweetDialog, self).__init__(parent)
        self.setWindowTitle("Tweet info")
        self.resize(300, 200)

    def show_tweet(self, tweet_data):
        id = QLabel()
        text = QLabel()
        position = QLabel()
        creation = QLabel()

        id.setText("ID: {0}".format(tweet_data.id))
        creation.setText("Create at: {0}".format(tweet_data.time))
        text.setText("Text: {0}".format(tweet_data.text))
        position.setText("Latitude: {0}, Longitude: {1}".format(tweet_data.position['lat'], tweet_data.position['lon']))

        layout = QVBoxLayout(self)

        layout.addWidget(id)
        layout.addWidget(creation)
        layout.addWidget(text)
        layout.addWidget(position)


class TweetProducer(Thread):
    def __init__(self, prediction_queue, data_queue):
        Thread.__init__(self)
        self.prediction_queue = prediction_queue
        self.data_queue = data_queue
        self.listWidget = None
        self.tweets = {}

    def run(self):
        # Create an PyQT4 application object.
        app = QApplication(sys.argv)

        # The QWidget widget is the base class of all user interface objects in PyQt4.
        w = QWidget()

        # Set window size.
        w.resize(400, 500)

        # Set window title
        w.setWindowTitle("Predicted Tweet")

        grid = QGridLayout(w)
        # Create a button in the window
        locate_btn = QPushButton('Locate tweet')
        locate_all_btn = QPushButton('Locate all')
        self.listWidget = TweetList()

        # connect the signals to the sqlots
        locate_btn.clicked.connect(self.on_click_one)
        locate_all_btn.clicked.connect(self.on_click_all)

        grid.addWidget(self.listWidget)
        grid.addWidget(locate_btn)
        grid.addWidget(locate_all_btn)

        # Show window
        w.show()
        app.exec_()

    def close_app(self):
        QCoreApplication.quit()

    def on_click_one(self):
        item = self.listWidget.currentItem()
        if item:
            item_id = item.text().split(': ')
            int_id = int(item_id[2])
            self.data_queue.put(self.tweets[int_id])

    def on_click_all(self):
        tweet_list = [tweet for tweet in self.tweets.values()]
        self.data_queue.put(tweet_list)

    def add_item_to_list(self, tweet_item):
        self.listWidget.update_tweets_list(tweet_item)
        self.tweets[int(tweet_item.id)] = tweet_item
        if self.listWidget.count() > 10:
            self.listWidget.takeItem(self.listWidget.count() - 1)
        if len(tweet_item.text) > 40:
            self.listWidget.insertItem(0, 'Text: ' + tweet_item.text[:35] + '...\nId: ' + tweet_item.id)
        else:
            self.listWidget.insertItem(0, 'Text: ' + tweet_item.text + '\nId: ' + tweet_item.id)


class MapConsumer(Thread):
    def __init__(self, prediction_queue, data_queue, folder):
        Thread.__init__(self)
        self.prediction_queue = prediction_queue
        self.data_queue = data_queue
        self.folder = folder

    def run(self):
        geoplotlib.tiles_provider({
            'url': lambda zoom, xtile, ytile: 'http://a.tile.openstreetmap.org/%d/%d/%d.png' % (zoom, xtile, ytile),
            'tiles_dir': 'a_my_tile',
            'attribution': 'Open Street Map'
        })
        geoplotlib.add_layer(PredictionLayer(bw=10, data_q=self.data_queue, folder=self.folder))
        geoplotlib.set_bbox(BoundingBox.from_points([6.911300, 18.136126], [36.406382, 47.363242]))

        geoplotlib.show()


class PredictionLayer(BaseLayer):
    def __init__(self, bw, cmap='hot', scaling='sqrt', alpha=220,
                 cut_below=None, clip_above=None, binsize=1, cmap_levels=20, data_q=None, folder=None):
        self.bw = bw
        self.folder = folder + '/'
        self.cmap = colors.ColorMap(cmap, alpha=alpha, levels=cmap_levels)
        self.scaling = scaling
        self.cut_below = cut_below
        self.clip_above = clip_above
        self.binsize = binsize
        self.data_q = data_q
        self.tweets = None
        self.tweet = None
        self.sq_id = -1
        self.loc_sq = ''
        self.loc_coordinates = {}
        refer_file = glob.glob(self.folder + "Visualization_data/Squares_data/*")
        ts_files_list = sorted(glob.glob(refer_file[0] + "/*"))
        self.ts_files = []
        for ts_file in ts_files_list:
            dir_name = os.path.basename(ts_file)
            if not dir_name == "columns.txt" and not dir_name == "rows.txt":
                self.ts_files.append(dir_name)
        self.current_file = 0

        path_to_locs = glob.glob(self.folder + "Visualization_data/Squares_data/*")
        loc_names = []
        for path in path_to_locs:
            loc_names.append(os.path.basename(path))
        for loc_name in loc_names:
            cols = {}
            with open(self.folder + "Visualization_data/Squares_data/" + str(loc_name) + "/columns.txt") as col_file:
                for line in col_file:
                    tokens = line.split("\t")
                    id = int(tokens[0])
                    sx = float(tokens[1])
                    dx = float(tokens[2])
                    cols.update({id: {"sx": sx, "dx": dx}})
            rows = {}
            with open(self.folder + "Visualization_data/Squares_data/" + str(loc_name) + "/rows.txt") as row_file:
                for line in row_file:
                    tokens = line.split("\t")
                    id = int(tokens[0])
                    up = float(tokens[1])
                    dw = float(tokens[2])
                    rows.update({id: {"up": up, "dw": dw}})
            self.loc_coordinates.update({loc_name: [cols, rows]})
        self.hotspots = HotspotManager()
        self.keywords = {}
        self.update_keywords()

    def draw(self, proj, mouse_x, mouse_y, ui_manager):
        self.painter = BatchPainter()
        lon = []
        lat = []
        if self.data_q.qsize() > 0:
            tmp = self.data_q.get()
            if type(tmp) is list:
                self.sq_id = -1
                self.tweet = None
                self.tweets = tmp
                for tweet in self.tweets:
                    lon.append(tweet.position['lon'])
                    lat.append(tweet.position['lat'])
                proj.fit(BoundingBox.from_points(lons=np.array(lon),
                                                 lats=np.array(lat)))
                xvt, yvt = proj.lonlat_to_screen(np.array(lon), np.array(lat))
                # self.painter.set_color([0, 0, 0])
                self.painter.points(xvt, yvt, rounded=True, point_size=10)
            elif isinstance(tmp, utilities.Tweet):
                self.tweets = None
                self.tweet = tmp
                lon.append(self.tweet.position['lon'])
                lat.append(self.tweet.position['lat'])
                xvt, yvt = proj.lonlat_to_screen(np.array(lon), np.array(lat))
                self.painter.circle_filled(xvt, yvt, r=10)

        elif self.tweets:
            self.sq_id = -1
            for tweet in self.tweets:
                lon.append(tweet.position['lon'])
                lat.append(tweet.position['lat'])
            xvt, yvt = proj.lonlat_to_screen(np.array(lon), np.array(lat))
            # self.painter.set_color([0, 0, 0])
            self.painter.points(xvt, yvt, rounded=True, point_size=15)
        elif self.tweet:
            lon_t = []
            lat_t = []
            for loc_name, loc_sq_ids in self.loc_coordinates.items():
                for c_id, c_limits in loc_sq_ids[0].items():
                    for r_id, r_limits in loc_sq_ids[1].items():
                        if (c_limits['sx'] <= self.tweet.position['lon'] < c_limits['dx']) and (
                                        r_limits['dw'] <= self.tweet.position['lat'] < r_limits['up']):
                            self.loc_sq = loc_name
                            sq_id = utilities.index_to_id(r_id, c_id)
                            self.sq_id = sq_id
                            lon = np.array([c_limits['sx'], c_limits['dx']])
                            lat = np.array([r_limits['dw'], r_limits['up']])
                            x, y = proj.lonlat_to_screen(lon, lat)
                            w = x.max() - x.min()
                            h = y.max() - y.min()
                            w = np.ceil(w / 2) * 2
                            h = np.ceil(h / 2) * 2
                            l = max(w, h)
                            n_x = [x.min(), x.min() + l, x.min() + l, x.min()]
                            n_y = [y.min() + l, y.min() + l, y.min(), y.min()]
                            self.painter.set_color([0, 0, 0])
                            self.painter.linestrip(n_x, n_y, closed=True)
                            self.hotspots.add_poly(np.array(n_x), np.array(n_y), self.keywords[loc_name][sq_id])
                            self.painter.set_color('b')
                            lon_t.append(self.tweet.position['lon'])
                            lat_t.append(self.tweet.position['lat'])
                            xvt, yvt = proj.lonlat_to_screen(np.array(lon_t), np.array(lat_t))
                            self.painter.circle_filled(xvt, yvt, r=10)
                            picked = self.hotspots.pick(mouse_x, mouse_y)
                            if picked:
                                ui_manager.tooltip(picked)
        self.painter.batch_draw()

    def on_mouse_release(self, proj, mouse_x, mouse_y, modifiers):
        if self.sq_id == -1:
            return
        p_lat, p_lon = proj.screen_to_latlon(mouse_x, mouse_y)
        loc_sq_ids = self.loc_coordinates[self.loc_sq]
        for c_id, c_limits in loc_sq_ids[0].items():
            for r_id, r_limits in loc_sq_ids[1].items():
                if (c_limits['sx'] <= p_lon < c_limits['dx']) and (r_limits['dw'] <= p_lat < r_limits['up']):
                    sq_id = utilities.index_to_id(r_id, c_id)
                    if sq_id == self.sq_id:
                        if os.path.exists(
                                                                        self.folder + "Visualization_data/Squares_data/" + self.loc_sq + "/Time_series/" + str(
                                                self.sq_id) + ".png"):
                            img_stream = open(
                                self.folder + "Visualization_data/Squares_data/" + self.loc_sq + "/Time_series/" + str(
                                    self.sq_id) + ".png",
                                "rb")
                            image = pyglet.image.load('time_series.png', file=img_stream)
                            window = pyglet.window.Window(width=image.width, height=image.height)
                            window.set_caption("Time series")

                            @window.event
                            def on_draw():
                                window.clear()
                                image.blit(0, 0)

                            pyglet.app.run()

    def update_keywords(self):
        self.keywords = {}
        for loc_name in self.loc_coordinates.keys():
            self.keywords.update({loc_name: {}})
            sq_files = glob.glob(
                self.folder + "Visualization_data/Squares_data/" + loc_name + "/" + self.ts_files[
                    self.current_file] + "/*")
            for sq_file_path in sq_files:
                sq_id = os.path.basename(sq_file_path)
                id = int(sq_id[:-4])
                with open(sq_file_path, 'r') as sq_file:
                    for line in sq_file:
                        s_line = line.split("=")
                        if s_line[0] == "keywords":
                            keywords_vect = s_line[1].split(" ")
                            keywords = "keywords: "
                            for w in keywords_vect:
                                if not w == '':
                                    keywords += w + ", "
                            self.keywords[loc_name].update({id: keywords[:-2]})
                            break
