'''
    TweeLoc is created by

    Pavlos Paraskevopoulos:
        George Mason University
        LIPADE, Paris Descartes University
        pparaske@gmu.edu

    Giovanni Pellegrini:
        University of Trento
        giovanni.pellegrini@unitn.it

    Themis Palpanas:
        Paris Descartes University
        themis@mi.parisdescartes.fr

    Copyright Paraskevopoulos, Pellegrini, Palpanas.


    This software cannot be used for commercial purposes.

    You may use the software as is, only for research purposes.
    You should acknowledge the authors by referencing the following two papers (the first describes the method and algorithms; the second describes the system and visualizations):

    Pavlos Paraskevopoulos, Themis Palpanas.
    Where has this Tweet Come From? Fast and Fine-Grained Geolocalisation of Non-Geotagged Tweets. Social Network Analysis and Mining (SNAM) Journal 6(1), pp 1-16, 2016.
    http://www.mi.parisdescartes.fr/~themisp/publications/snam16.pdf

    Pavlos Paraskevopoulos, Giovanni Pellegrini, Themis Palpanas. TweeLoc: A System for Geolocalizing Tweets at Fine-Grain. International Conference on Data Mining (ICDM), New Orleans, LA, USA, November 2017.
    http://www.mi.parisdescartes.fr/~themisp/publications/icdm17-tweeloc.pdf
'''

import geoplotlib
import glob
import os
import utilities
import stream
import time
import warnings
import sys

warnings.filterwarnings("ignore")
from geoplotlib.utils import read_csv
from geoplotlib.layers import *
from communicationinterface import Communication
from customlayers.densitylayer import DensityLayer
from customlayers.testlayer import TestPointsLayer
from customlayers.heatmap import TrainHeatMap, PredictionHeatMap, DifferentialHeatMap
from utilities import Tweet
from PySide.QtGui import *


class BaseModel:
    def _retrieve_data(self, layer, timestamp):
        """
        Retrieve the data to visualize depending of the timeslot and the type
        of visualization needed. For each visualization type, the function
        calls the appropriate subfunction
        :param layer: for which layer the data is retrieved
        :param timestamp: date and time of the file to open
        """
        if layer == "train":
            return self._retrieve_train_data(timestamp)
        elif layer == "test":
            return self._retrieve_test_data(timestamp)
        else:
            print "Layer " + layer + " Not implemented yet"
            return

    def _retrieve_train_data(self, timestamp):
        data = read_csv(self.datapath + "/Visualization_data/Points_data/Train/train_" + timestamp + ".csv")
        return data

    def _retrieve_test_data(self, timestamp):
        data = read_csv(self.datapath + "/Visualization_data/Points_data/Test/test_" + timestamp + ".csv")
        return data

    # Retrieve the list of the timeslots of the run
    def _retrieve_timestamps(self):
        refer_file = glob.glob(self.datapath + "/Visualization_data/Squares_data/*")
        list = sorted(glob.glob(refer_file[0] + "/*"))
        timestamp_list = [os.path.basename(p) for p in list]
        return timestamp_list[:-3]

    def _retrieve_window_size(self):
        with open(self.datapath + "/Visualization_data/Run_info.txt") as fin:
            for line in fin:
                if line.startswith("minutes_per_window"):
                    s_line = line.strip().split(":")
                    return int(s_line[1])

    # Retrieve the information [keywords, train, prediction, differential] for every location and every square
    def _retrieve_square_info(self, timestamp, info):
        data = {}
        location_list = self._retrieve_locations_names()
        for l in location_list:
            data[l] = {}
            square_files = glob.glob(self.datapath + "/Visualization_data/Squares_data/" + l + "/" + timestamp + "/*")
            for s in square_files:
                base = os.path.basename(s)
                id = int(base[:-4])
                with open(s, 'r') as sq_file:
                    for line in sq_file:
                        text = line.split('=')
                        if text[0] == "keywords" and info == "keywords":
                            data[l][id] = text[1].strip()
                            break
                        elif text[0] == "train" and info == "train":
                            data[l][id] = int(text[1].strip())
                            break
                        elif text[0] == "test" and info == "prediction":
                            data[l][id] = int(text[1].strip())
                            break
                        elif text[0] == "differential" and info == "differential":
                            data[l][id] = int(float(text[1].strip()))
                            break
        return data

    # Retrieve the coordinates of the squares of all the first level locations
    def _retrieve_locations_coordinates(self, locations):
        coordinates = {}
        for location in locations:
            cols = []
            with open(self.datapath + "/Visualization_data/Squares_data/" + str(location) + "/columns.txt") as col_file:
                for line in col_file:
                    tokens = line.split("\t")
                    sx = float(tokens[1])
                    dx = float(tokens[2])
                    cols.append((sx, dx))
            rows = []
            with open(self.datapath + "/Visualization_data/Squares_data/" + str(location) + "/rows.txt") as row_file:
                for line in row_file:
                    tokens = line.split("\t")
                    up = float(tokens[1])
                    dw = float(tokens[2])
                    rows.append((up, dw))
            coordinates[location] = [cols, rows]
        return coordinates

    # Retrieve the list with the names of the first level location
    def _retrieve_locations_names(self):
        location_folders = glob.glob(self.datapath + "/Visualization_data/Squares_data/*")
        location_list = [os.path.basename(p) for p in location_folders]
        return location_list

    def _retrieve_predictions(self):
        tweets = []
        prediction_files = glob.glob(self.datapath + "/Predictions/*")
        for f in sorted(prediction_files):
            with open(f, 'r') as tweets_file:
                for line in tweets_file:
                    tweet_json = json.loads(line.strip())
                    id_str = tweet_json['id_str']
                    lang = tweet_json['lang']
                    time = utilities.parse_tweet_time(tweet_json['created_at'])
                    text = utilities.parse_tweet_text(tweet_json['text'])
                    position = utilities.parse_tweet_position(
                        tweet_json['coordinates'])

                    tweets.append(Tweet(id_str, lang, time, text, position))
        return tweets


class Model(BaseModel):
    def __init__(self, datapath):
        self.datapath = datapath
        self.communicator = Communication(self)
        self.timestamplist = self._retrieve_timestamps()
        self.timestampid = 0
        self.VIS = ["DE", "TR", "PR", "DI"]
        self.VIS_COUNTER = 0
        self.coordinates = self._retrieve_locations_coordinates(self._retrieve_locations_names())
        self.window_size = self._retrieve_window_size()

    def runapp(self):
        app = QApplication(sys.argv)
        myapp = CreditsDialog()
        myapp.show()
        myapp.exec_()
        app.quit()
        # QCoreApplication.quit()
        # Retrieve the train and test data, from Run-x/Visualization_data/Points_data/[Train, Test]/
        traindata = self._retrieve_data("train", self.timestamplist[self.timestampid])

        # Set the tiles provider, i.e. the type of the map to visualize
        geoplotlib.tiles_provider({
            'url': lambda zoom, xtile, ytile: 'http://a.tile.openstreetmap.org/%d/%d/%d.png' % (zoom, xtile, ytile),
            'tiles_dir': 'a_my_tile',
            'attribution': 'Open Street Map'
        })

        # Set the bounding box from data. The BB is the portion of the map showed at the very beginning
        geoplotlib.set_bbox(BoundingBox.from_points(lons=traindata['lon'], lats=traindata['lat']))
        # geoplotlib.set_bbox(BoundingBox.from_points([6.911300, 18.136126], [36.406382, 47.363242]))

        geoplotlib.set_smoothing(True)
        self.update_view()
        # Show the data
        geoplotlib.show()

    # Point the timestampid to the next timeslot
    def load_next_timeslot(self):
        self.timestampid = (self.timestampid + 1) % len(self.timestamplist)
        self.update_data()

    # Point the timestampid to the previous timeslot
    def load_prev_timeslot(self):
        self.timestampid = (self.timestampid - 1) % len(self.timestamplist)
        self.update_data()

    # Change the view when UP button is pressed
    def load_next_view(self):
        self.VIS_COUNTER = (self.VIS_COUNTER + 1) % len(self.VIS)
        self.update_view()

    # Change the view when DOWN button is pressed
    def load_prev_view(self):
        self.VIS_COUNTER = (self.VIS_COUNTER - 1) % len(self.VIS)
        self.update_view()

    # After a LEFT, RIGHT event must update data from another timeslot
    def update_data(self):
        for w in pyglet.app.windows:
            for l in w.geoplotlib_config.layers:
                if isinstance(l, TestPointsLayer):
                    l.update_data(self._retrieve_data("test", self.timestamplist[self.timestampid]))
                elif isinstance(l, DensityLayer):
                    l.update_data(self._retrieve_data("train", self.timestamplist[self.timestampid]))
                elif isinstance(l, TrainHeatMap):
                    keywords = self._retrieve_square_info(self.timestamplist[self.timestampid], info="keywords")
                    counts = self._retrieve_square_info(self.timestamplist[self.timestampid], info="train")
                    l.update_data(keywords, counts)
                elif isinstance(l, PredictionHeatMap):
                    keywords = self._retrieve_square_info(self.timestamplist[self.timestampid], info="keywords")
                    counts = self._retrieve_square_info(self.timestamplist[self.timestampid], info="prediction")
                    l.update_data(keywords, counts)
                elif isinstance(l, DifferentialHeatMap):
                    keywords = self._retrieve_square_info(self.timestamplist[self.timestampid], info="keywords")
                    counts = self._retrieve_square_info(self.timestamplist[self.timestampid], info="differential")
                    l.update_data(keywords, counts)

                l.invalidate(w.proj)

    # After UP, DOWN event type of view must be changed
    def update_view(self):
        if self.VIS[self.VIS_COUNTER] == "DE":
            geoplotlib.clear()

            testdata = self._retrieve_data("test", self.timestamplist[self.timestampid])
            geoplotlib.add_layer(TestPointsLayer(testdata, point_size=8, communicator=self.communicator))

            traindata = self._retrieve_data("train", self.timestamplist[self.timestampid])
            geoplotlib.add_layer(
                DensityLayer(traindata, bw=3, cut_below=1e-5, binsize=1, communicator=self.communicator))

        elif self.VIS[self.VIS_COUNTER] == "TR":
            geoplotlib.clear()

            keywords = self._retrieve_square_info(self.timestamplist[self.timestampid], info="keywords")
            counts = self._retrieve_square_info(self.timestamplist[self.timestampid], info="train")
            geoplotlib.add_layer(TrainHeatMap(keywords=keywords, counts=counts, coordinates=self.coordinates,
                                              communicator=self.communicator))

        elif self.VIS[self.VIS_COUNTER] == "PR":
            geoplotlib.clear()

            keywords = self._retrieve_square_info(self.timestamplist[self.timestampid], info="keywords")
            counts = self._retrieve_square_info(self.timestamplist[self.timestampid], info="prediction")
            geoplotlib.add_layer(PredictionHeatMap(keywords=keywords, counts=counts, coordinates=self.coordinates,
                                                   communicator=self.communicator))

        elif self.VIS[self.VIS_COUNTER] == "DI":
            geoplotlib.clear()

            keywords = self._retrieve_square_info(self.timestamplist[self.timestampid], info="keywords")
            counts = self._retrieve_square_info(self.timestamplist[self.timestampid], info="differential")
            geoplotlib.add_layer(DifferentialHeatMap(keywords=keywords, counts=counts, coordinates=self.coordinates,
                                                     communicator=self.communicator))

        for w in pyglet.app.windows:
            for l in w.geoplotlib_config.layers:
                l.invalidate(w.proj)

    # Reference of ui_manager passed from the communicator, to write the info
    # about the timeslot to the screen
    def add_infobar(self, ui_manager, view):
        if view == "DE":
            ui_manager.info('Tweets Density - Window size: {0} minutes, starts at {1} '.format(self.window_size,
                                                                                               self.timestamplist[
                                                                                                   self.timestampid]))
        elif view == "TR":
            ui_manager.info(
                'Train Tweets - Window size: {0} minutes, starts at {1} '.format(self.window_size, self.timestamplist[
                    self.timestampid]))
        elif view == "PR":
            ui_manager.info('Predicted Tweets - Window size: {0} minutes, starts at {1} '.format(self.window_size,
                                                                                                 self.timestamplist[
                                                                                                     self.timestampid]))
        elif view == "DI":
            ui_manager.info('Differential Heatmap - Window size: {0} minutes, starts at {1} '.format(self.window_size,
                                                                                                     self.timestamplist[
                                                                                                         self.timestampid]))

    def add_statusbar(self, ui_manager, text):
        ui_manager.status(text)

    def on_mouse_release(self, name, id, modifiers):
        if os.path.exists(self.datapath + "/Visualization_data/Squares_data/" + name + "/Time_series/" + str(
                id) + ".png"):
            img_stream = open(
                self.datapath + "/Visualization_data/Squares_data/" + name + "/Time_series/" + str(id) + ".png", "rb")
            image = pyglet.image.load('time_series.png', file=img_stream)
            window = pyglet.window.Window(width=image.width, height=image.height)
            window.set_caption("Time series")

            @window.event
            def on_draw():
                window.clear()
                image.blit(0, 0)

            try:
                pyglet.app.run()
            except:
                print "Not able to retrieve the time series for this timeslot."


class StreamModel(BaseModel):
    def __init__(self, datapath):
        self.datapath = datapath
        self.predictions = self._retrieve_predictions()

    def runapp(self):
        prediction_queue = Queue.Queue()
        data_queue = Queue.Queue()
        producer = stream.TweetProducer(prediction_queue=prediction_queue, data_queue=data_queue)
        consumer = stream.MapConsumer(prediction_queue=prediction_queue, data_queue=data_queue,
                                      folder=self.datapath)

        producer.start()
        consumer.start()
        time.sleep(2)
        for tweet in self.predictions:
            if consumer.isAlive():
                prediction_queue.put(tweet)
                try:
                    producer.add_item_to_list(tweet)
                except:
                    break
                time.sleep(2)
            else:
                break
        # while producer.isAlive():
        #	time.sleep(1)
        producer.close_app()
        producer.join()
        consumer.join()

    def _retrieve_predictions(self):
        tweets = []
        prediction_files = glob.glob(self.datapath + "/Predictions/*")
        for f in sorted(prediction_files):
            with open(f, 'r') as tweets_file:
                for line in tweets_file:
                    tweet_json = json.loads(line.strip())
                    id_str = tweet_json['id_str']
                    lang = tweet_json['lang']
                    time = utilities.parse_tweet_time(tweet_json['created_at'])
                    text = utilities.parse_tweet_text(tweet_json['text'])
                    position = utilities.parse_tweet_position(
                        tweet_json['coordinates'])

                    tweets.append(Tweet(id_str, lang, time, text, position))
        return tweets


class CreditsDialog(QDialog):
    def __init__(self, parent=None):
        QDialog.__init__(self)
        self.setWindowTitle("TweeLoc")
        self.resize(500, 200)

    def show(self):
        tweeloc = QLabel()

        tweeloc.setText("TweeLoc - Fine-grained tweet's geolocalization tool\n"
                        "Developed by: Giovanni Pellegrini, Pavlos Paraskevopoulos, Themis Palpanas\n"
                        "Partially developed in collaboration with JOL-SKIL Telecom Italia s.p.a.")
        layout = QVBoxLayout(self)

        layout.addWidget(tweeloc)


if __name__ == '__main__':
    try:
        if sys.argv[1] == 'stream':
            model = StreamModel(sys.argv[2])
            model.runapp()
        elif sys.argv[1] == 'normal':
            model = Model(sys.argv[2])
            model.runapp()
        else:
            print "Usage: visualization.py normal/stream path/to/folder/with/data"
    except Exception as e:
        print str(e)
